#include <malloc.h>
#include <stdio.h>
#include "..\plugin.hpp"
#include "cfplugin.h"

#define FAR_CLONE_FILE "FarCloneFile"
#define FAR_CLONE_MUTEX "FarCloneMutex"

PluginStartupInfo Info;
FARSTANDARDFUNCTIONS FSF;
char PluginRootKey[80];
BOOL IsOldFAR=TRUE;

#include "cf_editor.cpp"

struct FARPanelInfo
{
  char Path[MAX_PATH];
  int CurrentItem;
  int TopPanelItem;
  int ViewMode;
  int SortMode;
  int SortOrder;
  int ItemsNumber;
};

struct FARInfo
{
  DWORD flag;
  DWORD edit;
  FARPanelInfo panels[2];
};

struct SelInfo
{
  char LongName[MAX_PATH];
  char ShortName[14];
};

struct EditInfo
{
  char FileName[MAX_PATH];
};

struct Options
{
  char FarPath[MAX_PATH];
  long Timeout;
  BOOL CloneSelection;
  BOOL CloneEditors;
  BOOL CloneViewers;
  BOOL ShowPanels;
  BOOL ShowViewer;
  BOOL ShowEditor;
} Opt;

char *GetMsg(int MsgId)
{
  return(Info.GetMsg(Info.ModuleNumber, MsgId));
};

void WINAPI _export SetStartupInfo(const struct PluginStartupInfo *Info)
{
  memset(&::Info, 0, sizeof(::Info));
  memmove(&::Info, Info, (Info->StructSize > (int)sizeof(::Info))?sizeof(::Info):Info->StructSize);
  if(Info->StructSize > FAR165_INFO_SIZE)
  {
    IsOldFAR=FALSE;
    ::FSF=*Info->FSF;
    ::Info.FSF=&::FSF;
    strcpy(PluginRootKey,Info->RootKey);
    strcat(PluginRootKey,"\\CloneFar");

    //init opt
    Opt.FarPath[0]=0;
    Opt.Timeout=10;
    Opt.CloneSelection=TRUE;
    Opt.CloneEditors=FALSE;
    Opt.CloneViewers=FALSE;
    Opt.ShowPanels=TRUE;
    Opt.ShowViewer=FALSE;
    Opt.ShowEditor=FALSE;
    
    //load opt
    HKEY hKey;
    DWORD Type;
    DWORD DataSize=0;
    if ((RegOpenKeyEx(HKEY_CURRENT_USER,PluginRootKey,0,
           KEY_QUERY_VALUE,&hKey))==ERROR_SUCCESS)
    {
      DataSize=sizeof(Opt.FarPath);
      RegQueryValueEx(hKey,"FarPath",0,&Type,(LPBYTE)Opt.FarPath,&DataSize);
      DataSize=sizeof(Opt.Timeout);
      RegQueryValueEx(hKey,"Timeout",0,&Type,(LPBYTE)&Opt.Timeout,&DataSize);
      DataSize=sizeof(Opt.CloneSelection);
      RegQueryValueEx(hKey,"CloneSelection",0,&Type,(LPBYTE)&Opt.CloneSelection,&DataSize);
      DataSize=sizeof(Opt.CloneEditors);
      RegQueryValueEx(hKey,"CloneEditors",0,&Type,(LPBYTE)&Opt.CloneEditors,&DataSize);
      DataSize=sizeof(Opt.CloneViewers);
      RegQueryValueEx(hKey,"CloneViewers",0,&Type,(LPBYTE)&Opt.CloneViewers,&DataSize);

      DataSize=sizeof(Opt.ShowPanels);
      RegQueryValueEx(hKey,"ShowPanels",0,&Type,(LPBYTE)&Opt.ShowPanels,&DataSize);
      DataSize=sizeof(Opt.ShowViewer);
      RegQueryValueEx(hKey,"ShowViewer",0,&Type,(LPBYTE)&Opt.ShowViewer,&DataSize);
      DataSize=sizeof(Opt.ShowEditor);
      RegQueryValueEx(hKey,"ShowEditor",0,&Type,(LPBYTE)&Opt.ShowEditor,&DataSize);
      
      RegCloseKey(hKey);
    }
    OnLoad();
  }
}

void WINAPI _export GetPluginInfo(struct PluginInfo *Info)
{
  if(!IsOldFAR)
  {
    Info->StructSize=sizeof(*Info);
    Info->Flags=PF_PRELOAD;
    if(!Opt.ShowPanels)
      Info->Flags|=PF_DISABLEPANELS;
    if(Opt.ShowViewer)
      Info->Flags|=PF_VIEWER;
    if(Opt.ShowEditor)
      Info->Flags|=PF_EDITOR;
    Info->DiskMenuStringsNumber=0;
    static char *PluginMenuStrings[1];
    PluginMenuStrings[0]=GetMsg(mName);
    Info->PluginMenuStrings=PluginMenuStrings;
    Info->PluginMenuStringsNumber=sizeof(PluginMenuStrings)/sizeof(PluginMenuStrings[0]);
    Info->PluginConfigStrings=PluginMenuStrings;
    Info->PluginConfigStringsNumber=sizeof(PluginMenuStrings)/sizeof(PluginMenuStrings[0]);
    Info->CommandPrefix="clone";
  }
}

struct InitDialogItem
{
  int Type;
  int X1, Y1, X2, Y2;
  int Focus;
  int Selected;
  unsigned int Flags;
  int DefaultButton;
  char *Data;
};

static void InitDialogItems(InitDialogItem *Init,FarDialogItem *Item,int ItemsNumber)
{
  for (int i=0;i<ItemsNumber;i++)
  {
    Item[i].Type=Init[i].Type;
    Item[i].X1=Init[i].X1;
    Item[i].Y1=Init[i].Y1;
    Item[i].X2=Init[i].X2;
    Item[i].Y2=Init[i].Y2;
    Item[i].Focus=Init[i].Focus;
    Item[i].Selected=Init[i].Selected;
    Item[i].Flags=Init[i].Flags;
    Item[i].DefaultButton=Init[i].DefaultButton;
    if ((unsigned)Init[i].Data<2000)
      strcpy(Item[i].Data,GetMsg((unsigned int)Init[i].Data));
    else
      strcpy(Item[i].Data,Init[i].Data);
  }
}

long WINAPI ConfigDialogProc(HANDLE hDlg,int Msg,int Param1,long Param2)
{
  switch(Msg)
  {
    case DN_INITDIALOG:
      Info.SendDlgMessage(hDlg,DM_SETTEXTLENGTH,2,MAX_PATH-1);
      break;
  }
  return Info.DefDlgProc(hDlg,Msg,Param1,Param2);
}

int WINAPI _export Configure(int ItemNumber)
{
  switch(ItemNumber)
  {
    case 0:
    {
      //Show dialog
      /*
        00000000001111111111222222222233333333334444444444555555
        01234567890123456789012345678901234567890123456789012345
      00                                                        00
      01    Clone FAR ͻ   01
      02    FAR Path:                                         02
      03        03
      04                                                      04
      05    10s Timeout                                       05
      06    [ ] Clone selection                               06
      07    [ ] Clone editors                                 07
      08    [ ] Treat editors as viewers                      08
      09    [ ] Show in panels                                08
      10    [ ] Show in viewer                                08
      11    [ ] Show in editor                                08
      12   Ķ   09
      13                    [ Save ]      [ Cancel ]          10
      14   ͼ   11
      15                                                        12
        00000000001111111111222222222233333333334444444444555555
        01234567890123456789012345678901234567890123456789012345
      */
     
      static struct InitDialogItem InitItems[]={
      /* 0*/  {DI_DOUBLEBOX,3,1,52,14,0,0,0,0,(char *)mName},
      /* 1*/  {DI_TEXT,5,2,0,0,0,0,0,0,(char *)mPath},
      /* 2*/  {DI_EDIT,5,3,50,0,1,0,0,0,""},
      /* 3*/  {DI_FIXEDIT,5,5,6,0,1,(int)"#9",0,0,""},
      /* 4*/  {DI_TEXT,7,5,0,0,0,0,0,0,(char *)mSec},
      /* 5*/  {DI_TEXT,9,5,0,0,0,0,0,0,(char *)mTimeout},
      /* 6*/  {DI_CHECKBOX,5, 6,0,0,0,0,0,0,(char *)mSelection},
      /* 7*/  {DI_CHECKBOX,5, 7,0,0,0,0,0,0,(char *)mEditor},
      /* 8*/  {DI_CHECKBOX,5, 8,0,0,0,0,0,0,(char *)mViewer},

      /* 9*/  {DI_CHECKBOX,5, 9,0,0,0,0,0,0,(char *)mShowPanel},
      /*10*/  {DI_CHECKBOX,5,10,0,0,0,0,0,0,(char *)mShowViewer},
      /*11*/  {DI_CHECKBOX,5,11,0,0,0,0,0,0,(char *)mShowEditor},

      /*12*/  {DI_TEXT,-1,12,0,0,0,0,DIF_SEPARATOR,0,""},
      /*13*/  {DI_BUTTON,0,13,0,0,0,0,DIF_CENTERGROUP,1,(char *)mSave},
      /*14*/  {DI_BUTTON,0,13,0,0,0,0,DIF_CENTERGROUP,0,(char *)mCancel}
      };

      struct FarDialogItem DialogItems[sizeof(InitItems)/sizeof(InitItems[0])];
      InitDialogItems(InitItems,DialogItems,sizeof(InitItems)/sizeof(InitItems[0]));
      strcpy(DialogItems[2].Data,Opt.FarPath);
      ltoa(Opt.Timeout,DialogItems[3].Data,10);
      DialogItems[6].Selected=Opt.CloneSelection;
      DialogItems[7].Selected=Opt.CloneEditors;
      DialogItems[8].Selected=Opt.CloneViewers;

      DialogItems[9].Selected=Opt.ShowPanels;
      DialogItems[10].Selected=Opt.ShowViewer;
      DialogItems[11].Selected=Opt.ShowEditor;
      
      int DlgCode=Info.DialogEx(Info.ModuleNumber,-1,-1,56,16,"Config",DialogItems,(sizeof(DialogItems)/sizeof(DialogItems[0])),0,0,ConfigDialogProc,0);
      if (DlgCode!=13)
        return(FALSE);

      strcpy(Opt.FarPath,DialogItems[2].Data);
      Opt.Timeout=atol(DialogItems[3].Data);
      if(Opt.Timeout<1) Opt.Timeout=10;
      Opt.CloneSelection=DialogItems[6].Selected;
      Opt.CloneEditors=DialogItems[7].Selected;
      Opt.CloneViewers=DialogItems[8].Selected;

      Opt.ShowPanels=DialogItems[9].Selected;
      Opt.ShowViewer=DialogItems[10].Selected ;
      Opt.ShowEditor=DialogItems[11].Selected;

      HKEY hKey;
      DWORD Disposition;
      if((RegCreateKeyEx(HKEY_CURRENT_USER,PluginRootKey,0,NULL,0,KEY_WRITE,NULL,&hKey,&Disposition)) == ERROR_SUCCESS)
      {
        RegSetValueEx(hKey,"FarPath",0,REG_EXPAND_SZ,(LPBYTE)Opt.FarPath,sizeof(Opt.FarPath));
        RegSetValueEx(hKey,"Timeout",0,REG_DWORD,(LPBYTE)&Opt.Timeout,sizeof(Opt.Timeout));
        RegSetValueEx(hKey,"CloneSelection",0,REG_DWORD,(LPBYTE)&Opt.CloneSelection,sizeof(Opt.CloneSelection));
        RegSetValueEx(hKey,"CloneEditors",0,REG_DWORD,(LPBYTE)&Opt.CloneEditors,sizeof(Opt.CloneEditors));
        RegSetValueEx(hKey,"CloneViewers",0,REG_DWORD,(LPBYTE)&Opt.CloneViewers,sizeof(Opt.CloneViewers));

        RegSetValueEx(hKey,"ShowPanels",0,REG_DWORD,(LPBYTE)&Opt.ShowPanels,sizeof(Opt.ShowPanels));
        RegSetValueEx(hKey,"ShowViewer",0,REG_DWORD,(LPBYTE)&Opt.ShowViewer,sizeof(Opt.ShowViewer));
        RegSetValueEx(hKey,"ShowEditor",0,REG_DWORD,(LPBYTE)&Opt.ShowEditor,sizeof(Opt.ShowEditor));
        RegCloseKey(hKey);
      }
      return(TRUE);
    }
  }
  return(FALSE);
}

HANDLE WINAPI _export OpenPlugin(int OpenFrom,int Item)
{
  switch(OpenFrom)
  {
    case OPEN_COMMANDLINE:
    {
      HANDLE mutex=OpenMutex(MUTEX_ALL_ACCESS,FALSE,FAR_CLONE_MUTEX);
      if(mutex)
      {
        WaitForSingleObject(mutex,INFINITE);
        HANDLE map=OpenFileMapping(FILE_MAP_READ|FILE_MAP_WRITE,FALSE,FAR_CLONE_FILE);
        if(map)
        {
          FARInfo *params=(FARInfo *)MapViewOfFile(map,FILE_MAP_READ|FILE_MAP_WRITE,0,0,0);
          if(params)
          {
            if(params->flag)
            {
              SelInfo *sel;
              EditInfo *edit;
              PanelRedrawInfo Redraw;
              PanelInfo PInfo;
              params->flag=FALSE;
              for(int i=0;i<2;i++)
              {
                Info.Control(INVALID_HANDLE_VALUE,FCTL_GETPANELINFO+i,&PInfo);
                Info.Control(INVALID_HANDLE_VALUE,FCTL_SETPANELDIR-9*i,params->panels[i].Path);
                Info.Control(INVALID_HANDLE_VALUE,FCTL_SETVIEWMODE+i,&(params->panels[i].ViewMode));
                if(PInfo.SortMode!=params->panels[i].SortMode)
                  Info.Control(INVALID_HANDLE_VALUE,FCTL_SETSORTMODE+i,&(params->panels[i].SortMode));
                Redraw.CurrentItem=params->panels[i].CurrentItem;
                Redraw.TopPanelItem=params->panels[i].TopPanelItem;
                Info.Control(INVALID_HANDLE_VALUE,FCTL_REDRAWPANEL+i,&Redraw);
                if(Opt.CloneSelection)
                {
                  //selection start
                  Info.Control(INVALID_HANDLE_VALUE,FCTL_GETPANELINFO+i,&PInfo);
                  sel=(SelInfo *)(params+1);
                  if(i) sel+=params->panels[0].ItemsNumber;
                  for(int j=0;j<PInfo.ItemsNumber;j++)
                  {
                    for(int k=0;k<params->panels[i].ItemsNumber;k++)
                      if(!(strcmp(PInfo.PanelItems[j].FindData.cFileName,sel[k].LongName)||strcmp(PInfo.PanelItems[j].FindData.cAlternateFileName,sel[k].ShortName)))
                      {
                        PInfo.PanelItems[j].Flags|=PPIF_SELECTED;
                        break;
                      }
                  }
                  Info.Control(INVALID_HANDLE_VALUE,FCTL_SETSELECTION+i,&PInfo);
                  Info.Control(INVALID_HANDLE_VALUE,FCTL_REDRAWPANEL+i,&Redraw);
                  //selection end
                }
              }
              if((params->edit)&&(Opt.CloneEditors))
              {
                sel=(SelInfo *)(params+1);
                sel+=params->panels[0].ItemsNumber+params->panels[1].ItemsNumber;
                edit=(EditInfo *)sel;
                for(unsigned int i=0;i<params->edit;i++)
                {
                  if(Opt.CloneViewers)
                    Info.Viewer(edit[i].FileName,edit[i].FileName,0,0,-1,-1,EF_NONMODAL);
                  else
                    Info.Editor(edit[i].FileName,edit[i].FileName,0,0,-1,-1,EF_NONMODAL,1,1);
                }
              }
            }
            UnmapViewOfFile(params);
          }
          CloseHandle(map);
        }
        ReleaseMutex(mutex);
        CloseHandle(mutex);
      }
      break;
    }
    case OPEN_EDITOR:
    case OPEN_VIEWER:
    case OPEN_PLUGINSMENU:
    {
      FARInfo fi;
      PanelInfo PInfo;
      SelInfo *sel[2]={NULL,NULL};
      EditInfo *edit=NULL;

      fi.flag=TRUE;
      for(int i=0;i<2;i++)
      {
        int selcount;
        Info.Control(INVALID_HANDLE_VALUE,FCTL_GETPANELINFO+i,&PInfo);
        strcpy(fi.panels[i].Path,PInfo.CurDir);
        fi.panels[i].CurrentItem=PInfo.CurrentItem;
        fi.panels[i].TopPanelItem=PInfo.TopPanelItem;
        fi.panels[i].ViewMode=PInfo.ViewMode;
        fi.panels[i].SortMode=PInfo.SortMode;
        selcount=PInfo.SelectedItemsNumber;
        if(selcount==1)
        {
          selcount=0;
          if(PInfo.SelectedItems[0].Flags&PPIF_SELECTED)
            selcount++;
        }
        fi.panels[i].ItemsNumber=selcount;
        if(fi.panels[i].ItemsNumber)
        {
          sel[i]=(SelInfo *)malloc(fi.panels[i].ItemsNumber*sizeof(SelInfo));
          if(sel[i])
            for(int j=0;j<fi.panels[i].ItemsNumber;j++)
            {
              strcpy(sel[i][j].LongName,PInfo.SelectedItems[j].FindData.cFileName);
              strcpy(sel[i][j].ShortName,PInfo.SelectedItems[j].FindData.cAlternateFileName);
            }
        }
        if(!sel[i]) fi.panels[i].ItemsNumber=0;
      }
      fi.edit=0;
      PEditFile curedit=editfiles->next;
      while(curedit!=editfiles)
      {
        fi.edit++;
        curedit=curedit->next;
      }
      if(fi.edit)
      {
        edit=(EditInfo *)malloc(sizeof(EditInfo)*fi.edit);
        curedit=editfiles->next;
        for(unsigned int i=0;i<fi.edit;i++,curedit=curedit->next)
          strcpy(edit[i].FileName,curedit->filename);
      }
      HANDLE map=CreateFileMapping(INVALID_HANDLE_VALUE,NULL,PAGE_READWRITE,0,sizeof(FARInfo)+(fi.panels[0].ItemsNumber+fi.panels[1].ItemsNumber)*sizeof(SelInfo)+fi.edit*sizeof(EditInfo),FAR_CLONE_FILE);
      if(map)
      {
        HANDLE mutex=CreateMutex(NULL,FALSE,FAR_CLONE_MUTEX);
        if(mutex!=NULL)
        {
          if(GetLastError()!=ERROR_ALREADY_EXISTS)
          {
            BYTE *params=(BYTE *)MapViewOfFile(map,FILE_MAP_READ|FILE_MAP_WRITE,0,0,0);
            if(params)
            {
              memcpy(params,&fi,sizeof(FARInfo));
              memcpy(params+sizeof(FARInfo),sel[0],fi.panels[0].ItemsNumber*sizeof(SelInfo));
              memcpy(params+sizeof(FARInfo)+fi.panels[0].ItemsNumber*sizeof(SelInfo),sel[1],fi.panels[1].ItemsNumber*sizeof(SelInfo));
              memcpy(params+sizeof(FARInfo)+(fi.panels[0].ItemsNumber+fi.panels[1].ItemsNumber)*sizeof(SelInfo),edit,fi.edit*sizeof(EditInfo));
              UnmapViewOfFile(params);
              char exp_path[MAX_PATH];
              ExpandEnvironmentStrings(Opt.FarPath,exp_path,MAX_PATH);
              if(ShellExecute(NULL,"open",exp_path,NULL,NULL,SW_SHOW)>(HINSTANCE)32)
              {
                int work=1;
                for(int i=0;(i<Opt.Timeout*2)&&work;i++)
                {
                  Sleep(500);
                  WaitForSingleObject(mutex,INFINITE);
                  FARInfo *params=(FARInfo *)MapViewOfFile(map,FILE_MAP_READ|FILE_MAP_WRITE,0,0,0);
                  if(params)
                  {
                    DWORD flag=params->flag;
                    UnmapViewOfFile(params);
                    if(!flag)
                      work=0;
                  }
                  else
                    work=0;
                  ReleaseMutex(mutex);
                }
              }
            }
          }
          CloseHandle(mutex);
        }
        CloseHandle(map);
      }
      break;
    }
  }
  return(INVALID_HANDLE_VALUE);
}

void WINAPI ExitFAR()
{
  if(!IsOldFAR)
    OnExit();
}

int WINAPI ProcessEditorEvent(int Event,void *Param)
{
  if(!IsOldFAR)
    return OnEditorEvent(Event,Param);
  return 0;
}
